import { BgColors, FontSizes, FontStyles, TextColors } from './style-types';
export declare enum Level {
    error = 1,
    warn = 2,
    log = 3,
    debug = 4,
}
/**
 * Something that looks like console.*. I suspect
 * that it might be useful to use a non-console thing
 * in tests.
 */
export interface Printer {
    error: typeof console.error;
    warn: typeof console.warn;
    log: typeof console.log;
    debug: typeof console.debug;
    dir: typeof console.dir;
}
export declare namespace Logger {
    interface Config {
        printer: Printer;
        env?: 'browser' | 'node';
    }
}
/**
 * A class which allows for colorful, tagged logging
 */
export declare class Logger {
    private level;
    private printer;
    private stylesInProgress;
    private prefixesAndStyles;
    private nonStringsToLog;
    private msgsAndStyles;
    private env;
    /**
     * Create a new instance of Logger
     * @param level the log level (1=error, 2=warn, etc...)
     * @param printer the object that actually prints messages (looks like console.*),
     */
    constructor(level?: number, printerOrConfig?: Printer | Logger.Config);
    /**
     * Adds a prefix that will be applied to all messages.
     * This is useful for indicating the context in which a message is logged,
     * without forcing developers to "re-state" that context over and over.
     *
     * @example
     * ```ts
     * const logger = new Logger();
     *
     * function fooFunction() {
     *   logger.pushPrefix('foo');
     *   ...
     *   logger.log('hello');
     *   logger.log('world');
     *   ...
     *    logger.popPrefix();
     * }
     *
     * logger.log('begin');   // "begin"
     * fooFunction();         // "[foo] hello"
     *                        // "[foo] world"
     * logger.log('end');     // "end"
     *
     * ```
     * @param name the name of the tag
     */
    pushPrefix(name: string): this;
    /**
     * Remove the most recently added prefix from the logger
     *
     * @see {Logger.pushPrefix}
     */
    popPrefix(): this;
    /**
     * Add custom CSS styles
     * @param style
     * a string of CSS, similar to what you'd use for <div style="">
     */
    css(style: string): this;
    /** Stage a string and accumulated styles for later console functions */
    txt(...args: any[]): this;
    error(...args: any[]): void;
    warn(...args: any[]): void;
    log(...args: any[]): void;
    debug(...args: any[]): void;
    /**
     * According to the logStyles in './style.ts', set up
     * a property for each, kind of like
     * ```ts
     *   {
     *     get red() { }, //  store "color: red"
     *     get bgYellow() { } // store "background-color: yellow"
     *   }
     * ```
     * Ultimately, we want to be able to do something like
     * logger.red.bgYellow.txt('Hello red and yellow')
     */
    private setupStyles();
    /**
     * Actually print the message to the printer (maybe console.*)
     * @param level the level of the current message
     * @param msg the message text
     */
    private printMessage(level);
    private printBrowserMessage(level);
    private printNodeMessage(level);
}
export declare type LoggerWithStyles = Logger & {
    [K in keyof (TextColors & BgColors & FontStyles & FontSizes)]: LoggerWithStyles;
};
export interface LoggerConstructor {
    new (level?: Level, printerOrConfig?: Printer | Logger.Config): LoggerWithStyles;
}
declare const _default: LoggerConstructor;
export default _default;
